import { objectKeys, objectLoop } from '@dzeio/object-util'

interface CacheItem<T = any> {
	value: T
	tags?: Array<string>
	ttl?: number
	created?: number
	deletable?: boolean
}

export type CacheItemOptions<T = any> = Omit<CacheItem<T>, 'created' | 'value'>

export default class Cache<T = any> {
	public config: {
		ttl: number | null
	} = {
		ttl: null
	}
	private global: Record<string, CacheItem<T>> = {}
	private scoped: Record<string, CacheItem<T>> = {}

	/**
	 * get the cache
	 * @param scope get the scoped elements
	 * @returns {Record<string, CacheItem<T>>}
	 */
	public getAll(scope: boolean = false) {
		if (scope) {
			return this.scoped
		} else {
			return this.global
		}
	}

	public set(key: string, value: T, scope?: string, options?: CacheItemOptions<T>): boolean {
		if (!key) {
			return false
		}
		const obj: CacheItem<T> = {
			...options,
			value
		}
		const ttl = (options?.ttl ?? this.config.ttl)
		if (typeof ttl === 'number') {
			obj.created = new Date().getTime() / 1000
			obj.ttl = ttl
		}
		if (scope) {
			this.scoped[`${scope}-${key}`] = obj
			return true
		}
		this.global[key] = obj
		return true
	}

	public update(key: string, scope?: string, patch: any = null): {
		updated: boolean
		exists: boolean
		value?: T
	} {
		const it = this.getItem(key, scope)
		if (!it) {
			return {
				updated: false,
				exists: false
			}
		}
		let value = it.value
		if (typeof value === 'number' && typeof patch === 'number') {
			(value as number) += patch
		} else if (typeof value === 'string') {
			value = patch
		} else if (typeof value === 'object') {
			value = {...value, ...patch}
		}
		it.value = value
		return {
			updated: true,
			exists: true,
			value: it.value
		}
	}

	public get(key: string, scope?: string): T | null {
		const obj: CacheItem<T> | undefined = this.getItem(key, scope)
		if (!obj) {
			return null
		}

		if (typeof obj.ttl === 'number') {
			const now = new Date().getTime() / 1000
			if (obj.ttl + (obj.created as number) < now) {
				this.delete(key, scope, true)
				return null
			}
		}

		return obj.value
	}

	public delete(key: string, scope?: string, force = false): {
		existed: boolean
		deleted: boolean
	} {
		const realKey = scope ? `${scope}-${key}` : key
		const store = scope ? this.scoped : this.global

		const exists = realKey in store
		if (!exists) {
			return {
				existed: false,
				deleted: true
			}
		}
		const it = store[realKey]
		if (!force && 'deletable' in it && !it.deletable) {
			return {
				existed: exists,
				deleted: false
			}
		}

		delete store[realKey]
		
		return {
			existed: exists,
			deleted: true
		}
	}

	public invalidate(tags: Array<string>): {
		global: Array<string>
		scoped: Array<[string, string]>
	} {
		const items: {
			global: Array<string>
			scoped: Array<[string, string]>
		} = {
			global: [],
			scoped: []
		}
		objectLoop(this.global, (value, key) => {
			if (!value.tags) {
				return
			}
			if (this.findMutuals(value.tags, tags)) {
				const res = this.delete(key)
				if (res.deleted) [
					items.global.push(key)
				]
			}
		})
		objectLoop(this.scoped, (value, fullKey) => {
			if (!value.tags) {
				return
			}
			if (this.findMutuals(value.tags, tags)) {
				const [scope, key] = fullKey.split('-', 1) as [string, string]
				const res = this.delete(key, scope)
				if (res.deleted) [
					items.scoped.push([scope, key])
				]
			}
		})
		return items
	}

	private findMutuals(arr1: Array<string>, arr2: Array<string>): Array<string> {
		return arr1.filter((it) => arr2.includes(it))
	}

	public length(scope?: true | string) {
		if (scope === true) {
			return objectKeys(this.global).length + objectKeys(this.scoped).length
		} else if (scope) {
			return objectKeys(this.scoped)
		}
		return objectKeys(this.global)
	}

	private getItem(key: string, scope?: string): CacheItem<T> | undefined {
		return scope ? this.scoped[`${scope}-${key}`] : this.global[key]
	}
}
