import express, { Request, Response, Router } from 'express'
import http from 'http'
import { Server } from 'ws'
import Cache, { CacheItemOptions } from './Cache'
const cache = new Cache()

const server = express()
const httpServer = http.createServer(server)
const wss = new Server({server: httpServer, path: '/ws'})
wss.on('connection', (ws) => {
	ws.on('message', (msg) => {
		let json: {type: 'get' | 'set'} & any
		try {
			json = JSON.parse(msg.toString())
		} catch {
			ws.send(JSON.stringify({error: 'invalid json received'}))
			return
		}
		if (json.type === 'get') {
			ws.send(JSON.stringify({
				value: cache.get(json.key, json.scope)
			}))
			return
		} else if (json.type === 'set') {
			const ok = cache.set(json.key, json.value, json.scope, json.options)
			ws.send(JSON.stringify({
				ok: ok
			}))
		} else {
			ws.send(JSON.stringify({
				error: 'json.type is not correctly set'
			}))
		}
	})
})

function set(req: Request, res: Response, key: string, scope?: string) {
	const body = req.body as {value: any, options?: CacheItemOptions}
	const ok = cache.set(key, body.value, scope, body.options)
	res.status(200).json({ ok })
}

function get(res: Response, key: string, scope?: string) {
	const item = cache.get(key, scope)
	if (!item) {
		res.status(404).end()
		return
	}
	/* It does Http as well */
	res.status(200).json({
		value: item
	})
}

function del(res: Response, key: string, scope?: string) {
	const item = cache.delete(key, scope)
	/* It does Http as well */
	res.status(item.deleted ? 200 : 400).json(item)
}

server
	.use(express.json())
	.post('/invalidate', (req, res) => {
		const tags = req.body.tags
		if (!Array.isArray(tags)) {
			res.status(400).json({error: true, message: 'tags MUST be an array'})
			return
		}
		res.json(cache.invalidate(tags))
	})
	.use(
		'/config',
		Router()
			.get('/', (_, res) => {
				res.json(cache.config)
			})
			.patch('/', (req, res) => {
				Object.assign(cache.config, req.body)
				res.json({ ok: true })
			})
	)
	.use(
		'/cache',
		Router()
			.put('/:key', (req, res) => set(req, res, req.params.key))
			.get('/:key', (req, res) => get(res, req.params.key))
			.delete('/:key', (req, res) => del(res, req.params.key))
			.patch('/:key', (req, res) => {
				res.json(cache.update(req.params.key, undefined, req.body.value))
			})
			.put('/:scope/:key', (req, res) => set(req, res, req.params.key, req.params.scope))
			.patch('/:scope/:key', (req, res) => {
				res.json(cache.update(req.params.key, req.params.scope, req.body.value))
			})
			.get('/:scope/:key', (req, res) => get(res, req.params.key, req.params.scope))
			.delete('/:scope/:key', (req, res) => del(res, req.params.key, req.params.scope))
	)

httpServer.listen(3000, () => {
	console.log(`Listening to port ${3000}`);
})